#include "AssetManager.h"
#include "Log.h"
#define STBI_NO_STDIO
#define STB_IMAGE_IMPLEMENTATION
#include "stb_image.h"

#define STB_TRUETYPE_IMPLEMENTATION
#include "stb_truetype.h"

AssetManager* AssetManager::ptr = nullptr;

bool AssetManager::loadTexture(const char* path, Texture* p_texture)
{
    FileBuffer texture_data = {};
    if (!loadFile(path, &texture_data))
        return false;

    int width, height, channels;
    const auto* raw_data = reinterpret_cast<const stbi_uc*>(texture_data.data);
    stbi_uc* data = stbi_load_from_memory(raw_data, static_cast<int>(texture_data.size), &width, &height, &channels, 4);
    if (!data) {
        ALOGE("Failed to parse texture file %s", path);
        releaseFileBuffer(texture_data);
        return false;
    }
    *p_texture = Texture(width, height, data);

    return true;
}

bool AssetManager::loadFontData(const char* path, float char_height, FontData* p_font)
{
    FileBuffer ttf_data = {};
    if (!loadFile(path, &ttf_data))
        return false;

    static unsigned char bitmap_memory[STBTT_PH_VALUE * STBTT_PH_VALUE];
    FontData font_data = {};
    auto* data = reinterpret_cast<const unsigned char*>(ttf_data.data);
    stbtt_BakeFontBitmap(data,
                         0,
                         char_height,
                         bitmap_memory,
                         STBTT_PH_VALUE, STBTT_PH_VALUE,
                         32, 225, // latin-1 printable characters range
                         font_data.char_data);
    font_data.char_height = char_height;
    int off = stbtt_GetFontOffsetForIndex(data,0);
    if (off < 0) {
        releaseFileBuffer(ttf_data);
        return false;
    }
    stbtt_fontinfo info;
    if (!stbtt_InitFont(&info, data, off)) {
        releaseFileBuffer(ttf_data);
        return false;
    }
    font_data.scale_factor = stbtt_ScaleForPixelHeight(&info, char_height);
    stbtt_GetFontVMetrics(&info, &font_data.ascent, &font_data.descent, &font_data.line_gap);

    int x0, y0, x1, y1;
    stbtt_GetFontBoundingBox(&info, &x0, &y0, &x1, &y1);
    font_data.baseline_adjust = font_data.scale_factor * static_cast<float>(-y0);

    *p_font = font_data;

    releaseFileBuffer(ttf_data);

    return true;
}

bool AssetManager::loadFontBitmap(const char *path, float char_height, Texture* p_texture)
{
    FileBuffer ttf_data = {};
    if (!loadFile(path, &ttf_data))
        return false;
    static unsigned char bitmap_memory[STBTT_PH_VALUE * STBTT_PH_VALUE];
    stbtt_bakedchar char_data[225];
    stbtt_BakeFontBitmap(reinterpret_cast<const unsigned char*>(ttf_data.data),
                         0,
                         char_height,
                         bitmap_memory,
                         STBTT_PH_VALUE, STBTT_PH_VALUE,
                         32, 225, // latin-1 printable characters range
                         char_data);
    *p_texture = Texture(STBTT_PH_VALUE, STBTT_PH_VALUE, bitmap_memory, GL_R8);
    releaseFileBuffer(ttf_data);
    return true;
}
